/*
	This file contains all functions to implement double linkes lists
	for use in tftp access control list processing

	Compile this module with -DTEST to produce a self test program
*/ 

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "types.h"
#include "dll.h"
#include "access.h"

#ifdef MEMTRACE
#include "memtrace.h"
#else
#define xmalloc malloc
#define xfree free
#endif

/*
	set up the head/tail pointers for a virgin DLL chain
	the elements of the chain are assumed to be removed
	prior to calling this procedure
*/

VOID clearChain (dll_t *head)
{
    dll_t *tail = head+1;

    head->next = tail->next = tail;
    head->prev = tail->prev = head;
#ifdef DLL_MARK
    head->mark = tail->mark = 0;
#endif
}

/*
	delete all elements of a DLL chain
	the elements are assumed to be dynamically allocated
	and are thus 'free'ed.
*/

VOID deleteChain (dll_t *head)
{
    dll_t *ptr, *next, *tail = head+1;

    for (ptr = head->next; ptr != tail; ptr = next) {
        next = ptr->next;
        xfree (ptr);
    }

    clearChain(head);
}

/*
	initialize a DLL chain head.
	dynamically allocates the head/tail structs if the
	input argument is set to NULL.
*/

dll_t *initChain(dll_t *head)
{
    if (head == NULL)
        head = (dll_t *)xmalloc(2*sizeof(dll_t));

    if (head != NULL)
        clearChain(head);

    return head;
}


/*
	checks a DLL chain and returns TRUE
	if the chain is empty
*/

INT isEmptyChain(dll_t *head)
{
    dll_t *tail = head+1;

    return ((head->next == tail));
}

/*
	removes an element from a DLL chain.
	controlled by the 2nd (flag) argument, returns
	a pointer to the next (DLL_GETNEXT) or previous
	(DLL_GETPREV) element. DLL_GETPREV is the default,
	as with mostly used forward list scans, this is
	the correct mode. If DLL_FREE is set too, the element
	to be removed is considered object to free().
*/

dll_t *removeElement (dll_t *elem, INT flag)
{
    dll_t *rp;

    if ((flag & DLL_GETNEXT))
        rp = elem->next;
    else
        rp = elem->prev;
    elem->prev->next = elem->next;
    elem->next->prev = elem->prev;
    if ((flag & DLL_FREE))
        free (elem);
    else
        elem->next = elem->prev = NULL;
    return rp;
}

#ifdef DLL_MARK
/*
	removes all elements from a chain, that
	have the given 'mark' value. 
	these elements are NOT 'free'ed.
	returns the number of removed elements.
*/

VOID clearByMark (dll_t *head, INT mark)
{
    dll_t *ptr, *tail = head+1;

    for (ptr = head->next; ptr != tail; ) {
        if (ptr->mark == mark)
            ptr = removeElement(ptr, DLL_GETNEXT);
        else
            ptr = ptr->next;
    }
}

/*
	same as 'clearByMark but 'free'es the removed elements
*/

VOID deleteByMark (dll_t *head, INT mark)
{
    dll_t *ptr, *tail = head+1;

    for (ptr = head->next; ptr != tail; ) {
        if (ptr->mark == mark)
            ptr = removeElement(ptr, DLL_GETNEXT|DLL_FREE);
        else
            ptr = ptr->next;
    }
}
#endif

/*
	inserts an element at the head (top) of the DLL chain
*/

VOID insertAtHead (dll_t *head, dll_t *elem)
{
    elem->next = head->next;
    elem->prev = head;
    head->next = elem->next->prev = elem;
}

/*
	inserts an element at the tail (bottom) of the DLL chain
*/

VOID insertAtTail (dll_t *head, dll_t *elem)
{
    dll_t *tail = ++head;

    elem->prev = tail->prev;
    elem->next = tail;
    head->prev = elem->prev->next = elem;
}

/*
	inserts a new element after a particular element of the DLL chain
*/

VOID insertAfter (dll_t *where, dll_t *elem)
{
    elem->prev  = where;
    elem->next  = where->next;
    where->next = elem->next->prev = elem;
}

/*
	inserts a new element before a particular element of the DLL chain
*/

VOID insertBefore (dll_t *where, dll_t *elem)
{
    elem->next  = where;
    elem->prev  = where->prev;
    where->prev = elem->prev->next = elem;
}

/*
	exchanges the position of 2 elements in a DLL chain
*/

VOID exchangeElements (dll_t *e1, dll_t *e2)
{
    dll_t *p1, *p2, *n1, *n2;
/*
	save the pointers to the forward- and backward link
*/
    n1 = e1->next;
    p1 = e1->prev;
    n2 = e2->next;
    p2 = e2->prev;
/*
	check if both elements are adjacent to handle this special case
*/
    if (n1 == e2) {		/* consecutive elements e1 -> e2 */
        p1->next = e2;
        n2->prev = e1;
        e1->prev = e2;
        e2->next = e1;
        e1->next = n2;
        e2->prev = p1;
    }
    else if (p1 == e2) {	/* consecutive elements e2 -> e1 */
        p2->next = e1;
        n1->prev = e2;
        e2->prev = e1;
        e1->next = e2;
        e2->next = n1;
        e1->prev = p2;
    }
    else {			/* elements not adjacent */
        e1->next = n2;
        e1->prev = p2;
        e2->next = n1;
        e2->prev = p1;
        n1->prev = p1->next = e2;
        n2->prev = p2->next = e1;
    }
}

/*
	sorts the elements of a DLL chain by the value of
	the 'mark'. the  'flag' argument controls ascending
	(DLL_SORTASC, default) or descending (DLL_SORTDESC)
	mode of sort. sorting arlgorithm is 'sort by exchanging'.
*/

#ifdef DLL_MARK
VOID sortByMark (dll_t *head, INT flag)
{
    dll_t *ptr, *next, *tail = head+1;
    INT ex;

    do {
        ex = 0;
        for (ptr = head->next; ptr != tail; ptr = next) {
            if ((next = ptr->next) == tail)
                break;
            if (flag == DLL_SORTDESC) {
                if (ptr->mark >= next->mark)
                    continue;
            }
            else {
                if (ptr->mark <= next->mark) 
                    continue;
            }
            ex++;
            exchangeElements (ptr, next);
        }
    } while (ex);
}

/*
	get next element of a DLL chain that matches the 'mark'
	starting after element 'start'. with start == NULL, scan
	begins at head of list. no checks are made weather 'start'
	is a valid element of the list specified by 'head'
*/

dll_t *getByMark (dll_t *head, dll_t *start, INT mark)
{
    dll_t *ptr, *tail = head+1;

    if (start == NULL)
        start = head;

    for (ptr = start->next; ptr != tail; ptr = ptr->next)
        if (ptr->mark == mark)
            return ptr;

    return NULL;
}
#endif

/*
	get next element of a DLL chain starting after element 'start'. 
	with start == NULL, scan begins at head of list. 
	no checks are made weather 'start' is a valid element of 
	the list specified by 'head'
*/

dll_t *getNext (dll_t *head, dll_t *start)
{
    dll_t *tail = head+1;

    if (start == NULL)
        start = head;

    start = start->next;

    return start == tail ? NULL : start;
}

/*
	get previous element of a DLL chain starting before element 'start'. 
	with start == NULL, scan begins at tail of list. 
	no checks are made weather 'start' is a valid element of 
	the list specified by 'head'
*/

dll_t *getPrev (dll_t *head, dll_t *start)
{
    dll_t *tail = head+1;

    if (start == NULL)
        start = tail;

    start = start->prev;

    return start == head ? NULL : start;
}

/*
	adds an elment to the end of a DLL chain. the pattern element
	must be given by an VOID pointer and its size.
	new memory space is allocated for the new element and
	the values of the pattern element are copied over into
	the new one. the pattern element ist left unchanged.

	attention must be paid, if the pattern element contains
	self-relative pointers, which may produce wrong references
	if copied over to the newly allocated element.
*/

INT addPlain (dll_t *head, 
#ifdef DLL_MARK
              INT mark,
#endif
              VOID *data, size_t length)
{
    dll_t *new;

    if ((new = (dll_t *)xmalloc(sizeof(dll_t) + length - 1)) == NULL)
        return -1;

    memcpy(new->data, data, length);
#ifdef DLL_CTYPE
    new->ctype = CTYPE_PLAIN;
#endif
#ifdef DLL_MARK
    new->mark = mark;
#endif

    insertAtTail (head, new);

    return 0;
}

/*
	check DLL chain for occurence of a loop condition
	forward and backward reference loop checks are controlled
	by the 'flag' values DLL_CHECKPREV (forward loop check) and
	DLL_CHECKNEXT (backward reference loop check).
	only one type of check is performed per call to loopCheck.
*/

dll_t *loopCheck(dll_t *head, INT flag)
{
    dll_t *d1, *d2;

    if ((flag & DLL_CHECKNEXT)) {	/* check 'next' chain for backward loops */
        for (d1 = NULL; (d1 = getNext(head, d1)) != NULL;) {
            if (d1->next == d1)
                return d1;
            for (d2 = NULL; (d2 = getNext(head, d2)) != d1;) {
                if (d1->next == d2)
                    return d1;
            }
        }
    }
    else if ((flag & DLL_CHECKPREV)) {	/* check 'prev' chain for foreward loops */
        for (d1 = NULL; (d1 = getPrev(head, d1)) != NULL;) {
            if (d1->prev == d1)
                return d1;
            for (d2 = NULL; (d2 = getPrev(head, d2)) != d1;) {
                if (d1->prev == d2)
                    return d1;
            }
        }
    }
    return NULL;
}


static VOID dumpout (FILE *fd, dll_t *dp, INT i, INT atstart)
{
    fprintf(fd, "%c%4d: %-10p  Next: %-10p  Prev: %-10p"
#ifdef DLL_CTYPE
                    "   Type: %7d"
#endif
#ifdef DLL_MARK
                    "  Mark: %d"
#endif
                    "\n",
                    atstart ? '*' : ' ', i, dp, dp->next, dp->prev
#ifdef DLL_CTYPE
                    , dp->ctype
#endif
#ifdef DLL_MARK
                    , dp->mark
#endif
                    ); 
}

/*
	dumps the contents of a DLL chain to 'stderr'
	if 'flag' is DLL_SHOWHEAD, the linkage element's values
	are included in the dump. 'dumpbody' is the pointer to a VOID
	function that performs the dump of the 'data' (body-) section of
	a chain element. this function is called with 2 arguments:
	1.) a pointer to the current DLL element,
	2.) a pointer to the body section of the current element.
	if flag is 'ored' with DLL_LOOPCHECK, the function 'loopCheck'
	is called in forward and reverse mode and the dump is performed
	on only these elements that are located before the 1st reverse
	loop and after the last forward loop.
*/

VOID dumpChain(FILE *fd, dll_t *head, dll_t *start, INT flag, 
               INT bodyflag, VOID (* dumpbody)())
{

    dll_t *tail, *dp;
    INT i;
    
    if (fd == NULL)
        fd = stderr;

    tail = head + 1;

    if ((flag & DLL_SHOWHEAD)) {
        if (head == NULL)
            fprintf(fd, "Head : %-10p\n",head);
        else
            fprintf(fd, "Head : %-10p  Next: %-10p  Prev: %-10p\n",
                    head, head->next, head->prev);

        if (tail == NULL)
            fprintf(fd, "Tail : %-10p\n",tail);
        else
            fprintf(fd, "Tail : %-10p  Next: %-10p  Prev: %-10p\n",
                    tail, tail->next, tail->prev);

        if (start == NULL)
            fprintf(fd, "Start: %-10p\n",start);
        else
            fprintf(fd, "Start: %-10p  Next: %-10p  Prev: %-10p\n",
                    start, start->next, start->prev);
    }

    if (head == NULL) return;

    if ((flag & DLL_LOOPCHECK)) {
        dll_t *db, *df;

        db = loopCheck(head, DLL_CHECKNEXT);
        df = loopCheck(head, DLL_CHECKPREV);

        if (db != NULL) {
            if (df != NULL) {
                i = 0;
                dp = NULL;
                do {
                    dp = getNext(head, dp);
                    if ((flag & DLL_SHOWHEAD))
                        dumpout(fd, dp, ++i, (dp == start));
                    if (dumpbody != NULL)
                        (dumpbody)(fd, dp, bodyflag);
                } while (dp != db);
            }
            fprintf (fd, "Backward loop from %10p to %10p\n", db, db->next); 
        }
        if (df != NULL) {
            if (db != NULL) {
                i = 0;
                dp = NULL;
                do {
                    dp = getPrev(head, dp);
                    if ((flag & DLL_SHOWHEAD))
                        dumpout(fd, dp, --i, (dp == start));
                    if (dumpbody != NULL)
                        (dumpbody)(fd, dp, bodyflag);
                } while (dp != df);
            }
            fprintf (fd, "Forward loop from %10p to %10p\n", df, df->prev); 
        }
        if ((df != NULL) | (db != NULL))
            return;
    }

    i = 0;
    dp = NULL;
    while ((dp = getNext(head, dp)) != NULL) {
        if ((flag & DLL_SHOWHEAD))
            dumpout(fd, dp, ++i, (dp == start));
        if (dumpbody != NULL)
            (dumpbody)(fd, dp, bodyflag);
    }
}
        

#ifdef TEST

#include "util.h"
#include "dllapps.h"

#define DumpChain element=0;dumpChain(NULL,dp,NULL,DLL_SHOWHEAD|DLL_LOOPCHECK,(dp == NULL),dbody)

static INT element;

static char *regcerror(int err)
{
    static char errbuf[128];

    if (!regerror(err, NULL, errbuf, sizeof(errbuf)))
        return "Non REGEX error";
    else
        return errbuf;
}

static void dbody (FILE *fd, dll_t *dp, int flag) 
{
    acl_t *ap;
    struct in_addr ia;
    static CHAR *cnames[] = { "Unknown", "Plain", "RegEx", "ACL", "PID"};

    ++element;
    fprintf(fd, "%c%4d: %-10p  Next: %-10p  Prev: %-10p   Type: %-7s  Mark: %d",
            flag ? '*' : ' ', element, dp, dp->next, dp->prev, 
            (dp->ctype>0 && dp->ctype<4) ? cnames[dp->ctype] : cnames[0],
            dp->mark);
    switch (dp->ctype) {
        case CTYPE_PLAIN:
            fprintf(fd, "  Data: %d",(INT)dp->data[0]);
            break;
        case CTYPE_ACL:
            ap = (acl_t *)dp->data;
            fprintf(fd, "\nPerm: %4.4x",ap->perm);
            ia.s_addr = ap->addr;
            fprintf(fd, "  Addr: %s",inet_ntoa(ia));
            ia.s_addr = ap->mask;
            fprintf(fd, "  Mask: %s",inet_ntoa(ia));
            break;
        case CTYPE_REGEX:
            fprintf(fd, "\nRegex: [%s]",((dlr_t *)dp)->raw);
            break;
        case CTYPE_PID:
            fprintf(fd, "  PID: %d",(int)(((dlp_t *)dp)->pid));
            break;
    }
    printf("\n");
}


INT main (INT argc, CHAR **argv)
{
    dll_t *dp, *dx, dlist[6];
    static CHAR *relist[]= { "^file1\\....", "file[1-5]\\.dat", "[a-z].*\\.dat",
                              ".*file1.dat", "file.\\.dat", "nope" };
    static CHAR *mplist[]= { "%b.%E",        "%B.%H%M%S",   "%Dsub2/%B", 
                             NULL,           NULL,          NULL };
    static CHAR *cklist[] = { "file1.xxx", "file5.dat", "/any/other.dat", 
                              "/myfile1.dat", "file0.dat", "nope" };
    static CHAR *alist[] = { "1.2.3.4",         "1.2.3.9"  , "8.8.8.8", "1.2.9.9",     "2.2.2.2" };
    static CHAR *mlist[] = { "255.255.255.255", "0.0.0.255", "0.0.0.0", "0.0.255.255", "0.0.0.0" };
    static CHAR *xlist[] = { "1.2.3.4",         "1.2.3.5",   "8.8.8.8", "1.2.4.5",     "9.9.9.9" };
   
    INT i, rtc;
    struct in_addr ia, im;

    puts ("double linked list test program");

    puts ("Step 1: creating a list");

    if ((dp = initChain(NULL)) == NULL) {
        puts("Cannot create Chain");
        exit(1);
    }
    DumpChain;

    for (i=0; i<(sizeof(dlist)/sizeof(dll_t)); i++) {
        dlist[i].mark = 1 + i/3;
        dlist[i].data[0] = i;
        dlist[i].next = dlist[i].prev = NULL;
    }

    puts ("Step 2: inserting 2 elements at head of the list");

    insertAtHead(dp, &dlist[1]);
    insertAtHead(dp, &dlist[0]);
    DumpChain;

    puts ("Step 3: inserting 2 elements at tail of the list");

    insertAtTail(dp, &dlist[4]);
    insertAtTail(dp, &dlist[5]);
    DumpChain;

    puts ("Step 4: inserting 2 elements in the middle of the list");

    insertAfter(&dlist[1], &dlist[2]);
    insertBefore(&dlist[4], &dlist[3]);
    DumpChain;

    puts ("Step 5: exchanging elements");


    puts ("elements 0/1 (consecutive ascending)");
    exchangeElements(&dlist[0], &dlist[1]);
    DumpChain;
    puts ("elements 0/1 (consecutive descending)");
    exchangeElements(&dlist[0], &dlist[1]);
    DumpChain;

    puts ("elements 2/4 (non consecutive)");
    exchangeElements(&dlist[2], &dlist[4]);
    DumpChain;

    puts ("elements 3/5 (non consecutive at end of list)");
    exchangeElements(&dlist[3], &dlist[5]);
    DumpChain;

    puts ("Step 6: selecting by mark");

    for (dx = dp; dx != NULL;) {
        dx = getByMark(dp, dx, 2);
        if (dx != NULL) printf ("Selection 2 retrieves %p Data: %d\n", 
                                dx,(INT)dx->data[0]);
    }

    puts ("Step 7: sorting by mark");

    sortByMark (dp, DLL_SORTASC);
    DumpChain;

    puts ("Step 8: deleting 3 elements (1st, 3rd, last)");

    removeElement(&dlist[0], DLL_GETPREV);
    removeElement(&dlist[2], DLL_GETPREV);
    removeElement(&dlist[5], DLL_GETPREV);
    DumpChain;

    puts ("Step 9: deleting the rest of the list");

    clearChain(dp);
    DumpChain;

    puts ("Step 10: setting up a regex list");

    for (i=0; i<(sizeof(relist)/sizeof(CHAR *)); i++)
        if ((rtc = addRePattern ((dlr_t *)dp,i,relist[i],mplist[i],0,NULL))) {
            CHAR *msgtxt;
            if (rtc == -1)
                msgtxt = "no dynamic memory";
            else
                msgtxt = regcerror(rtc);
            printf("Cannot insert regex %d [%s]: %d:%s\n",
                   i,relist[i],rtc,msgtxt);
            exit(1);
        }
    DumpChain;

    puts ("Step 11: looking up by regex match");

    for (i=0; i<(sizeof(cklist)/sizeof(CHAR *)); i++) {
        dlr_t *dr = NULL;
        INT c = 0;
        while ((dr = findRePattern((dlr_t *)dp, dr, 0, cklist[i])) != NULL)
            printf("Looking up [%s] returns: %d [%s] (%s)->(%s)\n",
                   cklist[i],++c,relist[dr->mark], 
                   (mplist[dr->mark] == NULL) ? "" : mplist[dr->mark],
                   (mplist[dr->mark] == NULL) ? "" : 
                   xlateFileMap(cklist[i], mplist[dr->mark], NULL));
        if (!c)
            printf("Looking up [%s] returns: *not found*\n", cklist[i]);
    }

    puts ("Step 12: deleting list element by mark (3)");

    DumpChain;
    clearByMark (dp, 3);
    DumpChain;
    clearChain  (dp);

    puts ("Step 13: setting up an ip addr/mask list");

    for (i=0; i<(sizeof(alist)/sizeof(CHAR *)); i++)
        if (addACL ((dla_t *)dp, 1, i, alist[i], mlist[i], NULL)) {
            printf("Cannot insert ACL %d [%s/%s]\n",i,alist[i],mlist[i]);
            exit(1);
        }
    DumpChain;

    puts ("Step 14: looking up by ip address");

    for (i=0; i<(sizeof(xlist)/sizeof(CHAR *)); i++) {
        dla_t *da = NULL;
        INT c = 0;
        in_addr_t aa = (in_addr_t)inet_addr(xlist[i]);

        while ((da = findACL((dla_t *)dp, da, 1, aa)) != NULL) {
            ia.s_addr = da->acl.addr;
            im.s_addr = da->acl.mask;
            printf("Looking up [%s] returns: %d [%d]",
                   xlist[i], ++c, da->acl.perm);
            printf("  %-16s", inet_ntoa(ia));
            printf("%-16s\n", inet_ntoa(im));
        }

        if (!c)
            printf("Looking up [%s] returns: *not found*\n", xlist[i]);
    }

    return 0;
}

#endif /* TEST */
